==> catadjust-0.1.0.tar.gz: catadjust-0.1.0/catadjust/hazelt_adjust.py <==
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
from tqdm.auto import tqdm


class HazardELTAdjustment:
    """Adjust a catastrophe model location-level ELT with hazard values
    to match arbitrary target location-level hazard EEF curves as closely
    as possible by scaling event rates.
    """
    def __init__(self, elt_raw, loccol, eventcol, ratecol, hazcol):
        """Load raw location-level ELT and pre-process.

        Parameters
        ----------
        elt_raw : DataFrame
            Raw location-level ELT.
        loccol: str
            Name of column containing locationIDs.
        eventcol: str
            Name of column containing eventIDs.
        ratecol: str
            Name of column containing event rates.
        hazcol: str
            Name of column containing event-location hazard intensity.
        """

        # Load ELT to be adjusted and pre-process
        elt = elt_raw.astype({loccol: np.int64, eventcol: np.int64,
                              ratecol: np.float64, hazcol: np.float64}
                            ).drop_duplicates([loccol, eventcol]
                                             ).sort_values([loccol, hazcol], ascending=[True, False])
        self.loccol = loccol
        self.eventcol = eventcol
        self.ratecol = ratecol
        self.hazcol = hazcol
        self.elt = self.calc_eef(elt)
        m = self.elt.shape[0]

        # Sorted array of unique eventIDs
        eventIDs_unique = np.sort(self.elt[eventcol].unique())
        self.nevents = eventIDs_unique.size

        # Convert eventIDs in ELT to indices in event array
        self.loceventixs = np.searchsorted(eventIDs_unique, self.elt[eventcol])

        # Indices in ELT where location changes
        locbreaks = np.nonzero(np.diff(self.elt[loccol]))[0] + 1
        self.loc_slicers = np.stack([np.r_[0, locbreaks], np.r_[locbreaks, m]]).T

    def calc_eef(self, elt):
        """Calculate EEFs from a location-level ELT sorted by descending hazard.

        Parameters
        ----------
        elt : DataFrame
            Processed and sorted (in descending hazard intensity) location-level ELT.

        Returns
        -------
        elt : DataFrame
            Input ELT with additional EEF column.
        """

        elt['eef'] = elt.groupby(self.loccol, sort=False)[self.ratecol].transform(np.cumsum)
        return elt

    def adjust(self, eefs_targ, x0=None, min_rate=1e-12, tol=1e-4, niter=1000, alpha=0.001):
        """Adjust ELT to match location-level hazard curves.

        Parameters
        ----------
        eefs_targ : Series or ndarray
            Target EEFs in the same order as the processed ELT.
        x0 : Series or ndarray, optional
            Initial guess to use for rate adjustment.
        min_rate : float, optional
            Minimum allowable rate constraint.
        tol : float, optional
            Convergence criterion for cost function. Iteration stops
            once the cost function (mean square error) is less than this value.
        niter : int, optional
            Maximum number of iterations.
        alpha : float, optional
            Learning rate in Adam gradient descent algorithm.

        Returns
        -------
        elt_adj : DataFrame
            Adjusted ELT.
        res : dict
            Results dict.
        fs : ndarray
            Learning curve.
        """

        if x0 is None:
            x0 = self.elt.groupby(self.eventcol)[self.ratecol].mean().values
        else:
            x0 = np.array(x0)

        eefs_targ = np.array(eefs_targ)
        args = (eefs_targ,)
        res, fs = self._adam(self._cost, x0, args, alpha=alpha, niter=niter, tol=tol, amin=min_rate)
        elt_adj = self.elt.copy()
        elt_adj[self.ratecol] = res['x'][self.loceventixs]
        elt_adj = self.calc_eef(elt_adj)
        return elt_adj, res, fs

    def _cost(self, theta, eefs_targ):
        """Cost function for fitting an ELT to a target EEF by adjusting event rates.

        Parameters
        ----------
        theta : ndarray
            Rates to calculate cost function for, in unique eventID order.
        eefs_targ : ndarray
            Target EEFs for location-events in the same order as the
            pre-processed ELT.

        Returns
        -------
        cost : float
            Cost function evaluated at theta.
        cost_grad : ndarray
            Gradient of cost function.
        """

        # Calculate EEFs for each location by chunked cumulative sums
        eefs_pred = np.empty_like(eefs_targ)

        # Expand event rates to event-location rates
        rates = theta[self.loceventixs]
        for a, b in self.loc_slicers:
            eefs_pred[a:b] = rates[a:b].cumsum()

        # Calculate deltas and cost function for current parameters
        deltas = eefs_pred - eefs_targ
        cost = (deltas**2).mean()

        # Calculate gradient of cost function wrt to event rates
        grad_cost = np.zeros_like(theta)
        for a, b in self.loc_slicers:
            grad_cost[self.loceventixs[a:b]] += deltas[a:b][::-1].cumsum()[::-1]

        return cost, 2*grad_cost/deltas.size

    def _adam(self, fun, x0, args=(), alpha=0.001, beta1=0.9, beta2=0.999,
              eps=1e-8, niter=1000, tol=1e-6, amin=-np.inf, amax=np.inf):
        """Adaptive Moment Estimation gradient descent with weight clipping.

        Parameters
        ----------
        cost : function
            Cost function which returns cost and gradient.
        x0 : ndarray
            Initial values for optimisation.
        args : tuple, optional
            Arguments to be passed to cost function.
        alpha : float, optional
            Learning rate.
        beta1 : float, optional
            Exponential decay rate for gradient momentum.
        beta2 : float, optional
            Exponential decay rate for gradient variance.
        tol : float, optional
            Convergence criterion for cost function. Iteration stops
            once the cost function (mean square error) is less than this value.
        amin : float, optional
            Minimum value allowed for input values.
        amax : float, optional
            Maximum value allowed for input values.

        Returns
        -------
        cost : float
            Cost function evaluated at theta.
        cost_grad : ndarray
            Gradient of cost function.
        """

        x, m, v = x0, 0, 0
        fs = np.zeros(niter)

        pbar = tqdm(range(niter))
        for i in pbar:
            fs[i], grad = fun(x, *args)

            pbar.set_description(f'f={fs[i]:.2e}{">" if fs[i] > tol else "<="}{tol:.2e}')
            if fs[i] < tol:
                return dict(x=x, fun=fs[i], jac=grad, nit=i), fs

            # Estimates of first and second moment of gradient
            m = (1 - beta1)*grad + beta1*m
            v = (1 - beta2)*grad**2 + beta2*v

            # Bias correction
            mhat = m/(1 - beta1**(i+1))
            vhat = v/(1 - beta2**(i+1))

            # Update step
            x = x - alpha * mhat/(np.sqrt(vhat) + eps)

            # Weight clipping
            x = np.clip(x, amin, amax)

        f, grad = fun(x, *args)
        print('Warning: Iteration limit reached before cost function converged within tolerance')
        return dict(x=x, fun=f, jac=grad, nit=i), fs

==> catadjust-0.1.1.tar.gz: catadjust-0.1.1/catadjust/hazelt_adjust.py <==
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
from tqdm.auto import tqdm


class HazardELTAdjustment:
    """Adjust a catastrophe model location-level ELT with hazard values
    to match arbitrary target location-level hazard EEF curves as closely
    as possible by scaling event rates.
    """
    def __init__(self, elt_raw, loccol, eventcol, ratecol, hazcol):
        """Load raw location-level ELT and pre-process.

        Parameters
        ----------
        elt_raw : DataFrame
            Raw location-level ELT.
        loccol: str
            Name of column containing locationIDs.
        eventcol: str
            Name of column containing eventIDs.
        ratecol: str
            Name of column containing event rates.
        hazcol: str
            Name of column containing event-location hazard intensity.
        """

        # Load ELT to be adjusted and pre-process
        elt = elt_raw.astype({loccol: np.int64, eventcol: np.int64,
                              ratecol: np.float64, hazcol: np.float64}
                            ).drop_duplicates([loccol, eventcol]
                                             ).sort_values([loccol, hazcol],
                                                           ascending=[True, False]).dropna()
        self.loccol = loccol
        self.eventcol = eventcol
        self.ratecol = ratecol
        self.hazcol = hazcol
        self.elt = self.calc_eef(elt)
        m = self.elt.shape[0]

        # Sorted array of unique eventIDs
        eventIDs_unique = np.sort(self.elt[eventcol].unique())
        self.nevents = eventIDs_unique.size

        # Convert eventIDs in ELT to indices in event array
        self.loceventixs = np.searchsorted(eventIDs_unique, self.elt[eventcol])

        # Indices in ELT where location changes
        locbreaks = np.nonzero(np.diff(self.elt[loccol]))[0] + 1
        self.loc_slicers = np.stack([np.r_[0, locbreaks], np.r_[locbreaks, m]]).T

    def calc_eef(self, elt):
        """Calculate EEFs from a location-level ELT sorted by descending hazard.

        Parameters
        ----------
        elt : DataFrame
            Processed and sorted (in descending hazard intensity) location-level ELT.

        Returns
        -------
        elt : DataFrame
            Input ELT with additional EEF column.
        """

        elt['eef'] = elt.groupby(self.loccol, sort=False)[self.ratecol].transform(np.cumsum)
        return elt

    def adjust(self, eefs_targ, x0=None, min_rate=1e-12, tol=1e-4, niter=1000, alpha=0.001):
        """Adjust ELT to match location-level hazard curves.

        Parameters
        ----------
        eefs_targ : Series or ndarray
            Target EEFs in the same order as the processed ELT.
        x0 : Series or ndarray, optional
            Initial guess to use for rate adjustment.
        min_rate : float, optional
            Minimum allowable rate constraint.
        tol : float, optional
            Convergence criterion for cost function. Iteration stops
            once the cost function (mean square error) is less than this value.
        niter : int, optional
            Maximum number of iterations.
        alpha : float, optional
            Learning rate in Adam gradient descent algorithm.

        Returns
        -------
        elt_adj : DataFrame
            Adjusted ELT.
        res : dict
            Results dict.
        fs : ndarray
            Learning curve.
        """

        eefs_targ = np.array(eefs_targ)

        # Best initial guess for adjusted rates
        if x0 is None:
            eefs_loc = np.split(eefs_targ, self.loc_slicers[1:,0])
            rates0 = np.concatenate([np.r_[eef_loc[0], np.diff(eef_loc)]
                                     for eef_loc in eefs_loc])
            x0 = np.array(pd.DataFrame({self.eventcol: self.elt[self.eventcol].values,
                                        self.ratecol : rates0}
                                      ).groupby(self.eventcol)[self.ratecol].mean())
        else:
            x0 = np.array(x0)

        args = (eefs_targ,)
        res, fs = self._adam(self._cost, x0, args, alpha=alpha, niter=niter, tol=tol, amin=min_rate)
        elt_adj = self.elt.copy()
        elt_adj[self.ratecol] = res['x'][self.loceventixs]
        elt_adj = self.calc_eef(elt_adj)
        return elt_adj, res, fs

    def _cost(self, theta, eefs_targ):
        """Cost function for fitting an ELT to a target EEF by adjusting event rates.

        Parameters
        ----------
        theta : ndarray
            Rates to calculate cost function for, in unique eventID order.
        eefs_targ : ndarray
            Target EEFs for location-events in the same order as the
            pre-processed ELT.

        Returns
        -------
        cost : float
            Cost function evaluated at theta.
        cost_grad : ndarray
            Gradient of cost function.
        """

        # Calculate EEFs for each location by chunked cumulative sums
        eefs_pred = np.empty_like(eefs_targ)

        # Expand event rates to event-location rates
        rates = theta[self.loceventixs]
        for a, b in self.loc_slicers:
            eefs_pred[a:b] = rates[a:b].cumsum()

        # Calculate deltas and cost function for current parameters
        deltas = eefs_pred - eefs_targ
        cost = (deltas**2).mean()

        # Calculate gradient of cost function wrt to event rates
        grad_cost = np.zeros_like(theta)
        for a, b in self.loc_slicers:
            grad_cost[self.loceventixs[a:b]] += deltas[a:b][::-1].cumsum()[::-1]

        return cost, 2*grad_cost/deltas.size

    def _adam(self, fun, x0, args=(), alpha=0.001, beta1=0.9, beta2=0.999,
              eps=1e-8, niter=1000, tol=1e-6, amin=-np.inf, amax=np.inf):
        """Adaptive Moment Estimation gradient descent with weight clipping.

        Parameters
        ----------
        cost : function
            Cost function which returns cost and gradient.
        x0 : ndarray
            Initial values for optimisation.
        args : tuple, optional
            Arguments to be passed to cost function.
        alpha : float, optional
            Learning rate.
        beta1 : float, optional
            Exponential decay rate for gradient momentum.
        beta2 : float, optional
            Exponential decay rate for gradient variance.
        tol : float, optional
            Convergence criterion for cost function. Iteration stops
            once the cost function (mean square error) is less than this value.
        amin : float, optional
            Minimum value allowed for input values.
        amax : float, optional
            Maximum value allowed for input values.

        Returns
        -------
        cost : float
            Cost function evaluated at theta.
        cost_grad : ndarray
            Gradient of cost function.
        """

        x, m, v = x0, 0, 0
        fs = np.zeros(niter)

        pbar = tqdm(range(niter))
        for i in pbar:
            fs[i], grad = fun(x, *args)

            pbar.set_description(f'f={fs[i]:.2e}{">" if fs[i] > tol else "<="}{tol:.2e}')
            if fs[i] < tol:
                return dict(x=x, fun=fs[i], jac=grad, nit=i), fs

            # Estimates of first and second moment of gradient
            m = (1 - beta1)*grad + beta1*m
            v = (1 - beta2)*grad**2 + beta2*v

            # Bias correction
            mhat = m/(1 - beta1**(i+1))
            vhat = v/(1 - beta2**(i+1))

            # Update step
            x = x - alpha * mhat/(np.sqrt(vhat) + eps)

            # Weight clipping
            x = np.clip(x, amin, amax)

        f, grad = fun(x, *args)
        print('Warning: Iteration limit reached before cost function converged within tolerance')
        return dict(x=x, fun=f, jac=grad, nit=i), fs
